package apps;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.telelogic.rhapsody.core.IRPDependency;
import com.telelogic.rhapsody.core.IRPInstance;
import com.telelogic.rhapsody.core.IRPMatrixView;
import com.telelogic.rhapsody.core.IRPModelElement;
import com.telelogic.rhapsody.core.IRPTableLayout;
import com.telelogic.rhapsody.core.IRPTableView;

public class MBDGSimpleModelBuilder {
	private MBDGSimpleModel simpleModel;
	private boolean withDescendants = true;
	
	public MBDGSimpleModelBuilder(MBDGSimpleModel simpleModel) {
		setSimpleModel(simpleModel);
	}

	public void setWithDescendants(boolean withDescendants) {
		this.withDescendants = withDescendants;
	}	
	
	public void build(IRPModelElement modelElement) throws IOException {
		MBDGElement root = new MBDGElement(modelElement);
		List<MBDGElement> elements = cloneElementForEachModelReference(root);
		for (MBDGElement element : elements) {
			// Build tree of mbdg elements which represent report
			if (withDescendants) {
				build(element);				
			}
			root.addDescendent(element);
		}
		simpleModel.setRoot(root);
	}
	
	protected void build(MBDGElement element) throws IOException {
		List<IRPModelElement> nestedElements = element.getNestedElements();
		for (IRPModelElement nestedElement : nestedElements) {
			if (nestedElement instanceof IRPInstance) {
				IRPInstance part = (IRPInstance) nestedElement;	
				MBDGElement descendant2Clone = new MBDGElement(part);
				descendant2Clone.initScope(element);
				// Per each reference element create report element to represent it
				List<MBDGElement> descendants = cloneElementForEachModelReference(descendant2Clone);
				for (MBDGElement descendant : descendants) {
					element.addDescendent(descendant);
					// Build each descendant recursively
					build(descendant);
				}
			}
		}
	}
	
	protected List<MBDGElement> cloneElementForEachModelReference(MBDGElement element) throws IOException {
		List<MBDGElement> elements = new ArrayList<MBDGElement>();
		List<IRPModelElement> referencedModelElements = getReferencedModelElements(element);
		if (MBDGUtils.isType(element, "DocAbstractList") || MBDGUtils.isType(element, "DocAbstractTable")) {
			MBDGElement clonedElement = element.partialClone();
			clonedElement.setItems(referencedModelElements);
			elements.add(clonedElement);
		} else {
			for (IRPModelElement referencedModelElement : referencedModelElements) {
				MBDGElement clonedElement = element.partialClone();
				clonedElement.setReferencedModelElement(referencedModelElement);
				elements.add(clonedElement);
			}
			if (referencedModelElements.size() == 0 && !hasReferencedQuery(element)) {
				MBDGElement clonedElement = element.partialClone();
				elements.add(clonedElement);
			}
		}
		return elements;
	}

	protected List<IRPModelElement> getReferencedModelElements(MBDGElement element) throws IOException {
		List<IRPModelElement> referencedElements = new ArrayList<IRPModelElement>();
		if (element != null && element.getElement() != null) {
			List<IRPModelElement> allModelElements = getAllModelElements(element);
			for (IRPModelElement modelElement : allModelElements) {
				if (isQuery(modelElement)) {
					IRPModelElement query = modelElement;
					List<IRPModelElement> modelElements = getReferencedModelElements(query, element.getScope());
					referencedElements.addAll(modelElements);			
				} else {
					referencedElements.add(modelElement);
				}
			}
		}
		return referencedElements;
	}
	
	protected boolean isQuery(IRPModelElement modelElement) {
		return (modelElement instanceof IRPTableView || modelElement instanceof IRPTableLayout || modelElement instanceof IRPMatrixView);	
	}
	
	protected List<IRPModelElement> getAllModelElements(MBDGElement element) {
		List<IRPModelElement> docElementReferences = new ArrayList<IRPModelElement>();
		List<IRPModelElement> nestedElements = element.getNestedElements();
		for (IRPModelElement nestedElement : nestedElements) {
			if (nestedElement instanceof IRPDependency) {
				IRPDependency dependency = (IRPDependency)nestedElement;
				// Look for elements by dependency type
				if (MBDGUtils.isType(dependency, "DocElementReference")) {			
					IRPModelElement dependsOn = dependency.getDependsOn();
					// Document element can't be scope
					if (!MBDGUtils.isType(dependsOn, "DocElement")) {
						docElementReferences.add(dependsOn);
					}	
				}
			}
		}
		return docElementReferences;
	}	
	
	public List<IRPModelElement> getReferencedModelElements(IRPModelElement query, IRPModelElement scope) {
		List<IRPModelElement> modelElements = new ArrayList<IRPModelElement>();
		List<Object> elements = MBDGUtils.getResultList(query, scope);
		for (Object object : elements) {
			if (object instanceof IRPModelElement) {
				IRPModelElement modelElement = (IRPModelElement)object;
				if (modelElement.getName().equals("TopLevel")) { 
					continue;
				}
				modelElements.add(modelElement);
			}
		}
		return modelElements;
	}
	
	public boolean hasReferencedQuery(MBDGElement element) {
		boolean res = false;
		List<IRPModelElement> nestedElements = element.getNestedElements();
		for (IRPModelElement nestedElement : nestedElements) {
			if (nestedElement instanceof IRPDependency) {
				IRPDependency dependency = (IRPDependency)nestedElement;
				// Look for elements by dependency type
				if (MBDGUtils.isType(dependency, "DocElementReference")) {			
					IRPModelElement dependsOn = dependency.getDependsOn();
					if (isQuery(dependsOn)) {
						res = true;
						break;
					}
				}
			}
		}
		return res;
	}	

	public MBDGSimpleModel getSimpleModel() {
		return simpleModel;
	}

	public void setSimpleModel(MBDGSimpleModel simpleModel) {
		this.simpleModel = simpleModel;
	}
}
